BUFFER_SIZE = 100
ELEMENT_BUFFER_SIZE = 10

MAX_BBOXES_TO_CHECK_IN_PAST = 25
MAX_FRAMES_TO_CHECK_IN_PAST = 30

MIN_IOU_TO_CONSIDER_OVERLAP = 0.35
DIFFERENCE_TO_CONSIDER_ELEMENT_LEFT_SCENE_IN_SECONDS = 15.
MAX_RECOVER_TRACK_ATTEMPTS = 10

# HISTORY KEYS

INITIAL_CLASS_NAME = "initial_class_name"
OUT_CLASS_NAME = "out_class_name"
INITIAL_TIMESTAMP = "initial_timestamp"
OUT_TIMESTAMP = "out_timestamp"
FINITE_STATE_MACHINE = "finite_state_machine"
TRACK_LENGTH = "track_length"
OUT_BBOX_XYXY = "out_bbox_xyxy"
INITIAL_BBOX_XYXY = "initial_bbox_xyxy"
